var vueComposable = (function (exports, runtimeCore) {
    'use strict';

    function unwrap(o) {
        return runtimeCore.isRef(o) ? o.value : o;
    }
    function wrap(o) {
        return runtimeCore.isRef(o) ? o : runtimeCore.ref(o);
    }
    const isArray = Array.isArray;
    const isFunction = (val) => typeof val === "function";
    const isString = (val) => typeof val === "string";
    const isSymbol = (val) => typeof val === "symbol";
    const isBoolean = (val) => typeof val === "boolean";
    const isDate = (val) => isObject(val) && isFunction(val.getTime);
    const isNumber = (val) => typeof val === "number";
    const isObject = (val) => val !== null && typeof val === "object";
    const isElement = (val) => isObject(val) && !!val.tagName;
    function isPromise(val) {
        return isObject(val) && isFunction(val.then) && isFunction(val.catch);
    }
    const NO_OP = () => { };
    const FALSE_OP = () => false;
    const PASSIVE_EV = { passive: true };
    function promisedTimeout(timeout) {
        return new Promise(res => {
            setTimeout(res, timeout);
        });
    }
    function minMax(val, min, max) {
        if (val < min)
            return min;
        if (val > max)
            return max;
        return val;
    }
    const isClient = typeof window != 'undefined';
    // compact version: https://stackoverflow.com/a/33146982/1209882
    /**
     * returns a random string
     * @param len length of the string max: 36
     */
    // export function randomString(len: number) {
    //   return (+new Date).toString(36).slice(-len);
    // }

    function usePagination(options) {
        const _currentPage = wrap(options.currentPage);
        const _pageSize = wrap(options.pageSize);
        const _offset = runtimeCore.ref(0);
        const total = wrap(options.total);
        const offset = runtimeCore.computed({
            get() {
                return _offset.value;
            },
            set(v) {
                if (!isNumber(v)) {
                    return;
                }
                _offset.value = Math.min(v, total.value);
            }
        });
        const currentPage = runtimeCore.computed({
            get() {
                return _currentPage.value;
            },
            set(v) {
                if (!isNumber(v)) {
                    return;
                }
                _currentPage.value = minMax(v, 1, lastPage.value);
                // set the offset
                offset.value = (_currentPage.value - 1) * pageSize.value;
            }
        });
        const pageSize = runtimeCore.computed({
            get() {
                return _pageSize.value;
            },
            set(v) {
                if (!isNumber(v)) {
                    return;
                }
                _pageSize.value = v;
            }
        });
        const lastPage = runtimeCore.computed(() => Math.ceil(total.value / pageSize.value));
        // make sure the current page is the correct value
        currentPage.value = _currentPage.value;
        const prev = () => --currentPage.value;
        const next = () => ++currentPage.value;
        const first = () => (currentPage.value = 1);
        const last = () => (currentPage.value = lastPage.value);
        runtimeCore.watch([total, pageSize], () => {
            if (currentPage.value > lastPage.value) {
                currentPage.value = lastPage.value;
            }
        }, { lazy: true } // no need to run on first render
        );
        return {
            // Mutable state
            pageSize,
            total,
            currentPage,
            offset,
            // Computed
            lastPage,
            // Functions
            next,
            prev,
            first,
            last
        };
    }

    function useArrayPagination(array, options) {
        const arrayRef = wrap(array);
        const pagination = usePagination({
            ...{
                currentPage: 1,
                pageSize: 10,
            },
            ...options,
            total: runtimeCore.computed(() => arrayRef.value.length)
        });
        const result = runtimeCore.computed(() => {
            const array = arrayRef.value;
            if (!Array.isArray(array))
                return [];
            return array.slice(pagination.offset.value, pagination.offset.value + pagination.pageSize.value);
        });
        return {
            ...pagination,
            result
        };
    }

    function usePromise(fn, throwException = false) {
        if (!fn) {
            throw new Error(`[usePromise] argument can't be '${fn}'`);
        }
        if (typeof fn !== "function") {
            throw new Error(`[usePromise] expects function, but received ${typeof fn}`);
        }
        const loading = runtimeCore.ref(false);
        const error = runtimeCore.ref(null);
        const result = runtimeCore.ref(null);
        const promise = runtimeCore.ref();
        const exec = async (...args) => {
            loading.value = true;
            error.value = null;
            result.value = null;
            let throwExp = args && fn.length !== args.length && args.length > 0 && isBoolean(args[args.length - 1]) ? args[args.length - 1] : throwException;
            const currentPromise = (promise.value = fn(...args));
            try {
                const r = await currentPromise;
                if (promise.value === currentPromise) {
                    result.value = r;
                }
                return r;
            }
            catch (er) {
                if (promise.value === currentPromise) {
                    error.value = er;
                    result.value = null;
                }
                return throwExp ? currentPromise : undefined;
            }
            finally {
                if (promise.value === currentPromise) {
                    loading.value = false;
                }
            }
        };
        return {
            exec,
            result,
            promise,
            loading,
            error
        };
    }

    function useCancellablePromise(fn, throwException = false) {
        const cancelled = runtimeCore.ref(false);
        let _cancel = undefined;
        const cancel = (result) => _cancel(result); // TODO add warning if cancel is undefined
        const promise = (p) => new Promise((res, rej) => {
            _cancel = result => {
                cancelled.value = true;
                rej(result);
            };
            p.then(res).catch(rej);
        });
        const use = usePromise((...args) => promise(fn(...args)), throwException);
        return {
            ...use,
            cancel,
            cancelled
        };
    }

    const MAX_RETRIES = 9000;
    /* istanbul ignore next */
    const ExecutionId = Symbol( ``);
    /* istanbul ignore next */
    const CancellationToken = Symbol( ``);
    const defaultStrategy = async (options, context, factory, args) => {
        const retryId = context[ExecutionId].value;
        let i = -1;
        const maxRetries = options.maxRetries || MAX_RETRIES + 1;
        const delay = options.retryDelay || noDelay;
        context.retryErrors.value = [];
        context.isRetrying.value = false;
        context.nextRetry.value = undefined;
        let nextRetry = undefined;
        do {
            let success = false;
            let result = null;
            try {
                ++i;
                if (args) {
                    result = factory(...args);
                }
                else {
                    result = factory();
                }
                if (isPromise(result)) {
                    result = await result;
                }
                // is cancelled?
                if (context[CancellationToken].value) {
                    return null;
                }
                success = true;
            }
            catch (error) {
                result = null;
                context.retryErrors.value.push(error);
            }
            // is our retry current one?
            if (retryId !== context[ExecutionId].value) {
                return result;
            }
            if (success) {
                context.isRetrying.value = false;
                context.nextRetry.value = undefined;
                return result;
            }
            if (i >= maxRetries) {
                context.isRetrying.value = false;
                context.nextRetry.value = undefined;
                return Promise.reject(new Error(`[useRetry] max retries reached ${maxRetries}`));
            }
            context.isRetrying.value = true;
            const now = Date.now();
            const pDelayBy = delay(i); // wrapped
            const delayBy = isPromise(pDelayBy) ? await pDelayBy : pDelayBy; // unwrap promise
            if (!isPromise(pDelayBy) || !!delayBy) {
                if (isNumber(delayBy)) {
                    nextRetry = delayBy;
                }
                else if (isDate(delayBy)) {
                    nextRetry = delayBy.getTime();
                }
                else {
                    throw new Error(`[useRetry] invalid value received from options.retryDelay '${typeof delayBy}'`);
                }
                // if the retry is in the past, means we need to await that amount
                if (nextRetry < now) {
                    context.nextRetry.value = now + nextRetry;
                }
                else {
                    context.nextRetry.value = nextRetry;
                    nextRetry = nextRetry - now;
                }
                if (nextRetry > 0) {
                    await promisedTimeout(nextRetry);
                }
            }
            // is cancelled?
            if (context[CancellationToken].value) {
                return null;
            }
            // is our retry current one?
            if (retryId !== context[ExecutionId].value) {
                return result;
            }
        } while (i < MAX_RETRIES);
        return null;
    };
    function useRetry(options, factory) {
        const opt = !options || isFunction(options) ? {} : options;
        const fn = isFunction(options) ? options : factory;
        if (!isFunction(options) && !isObject(options)) {
            throw new Error("[useRetry] options needs to be 'object'");
        }
        if (!!fn && !isFunction(fn)) {
            throw new Error("[useRetry] factory needs to be 'function'");
        }
        const isRetrying = runtimeCore.ref(false);
        const nextRetry = runtimeCore.ref();
        const retryErrors = runtimeCore.ref([]);
        const cancellationToken = { value: false };
        const retryId = { value: 0 };
        const retryCount = runtimeCore.computed(() => retryErrors.value.length);
        const context = {
            isRetrying,
            retryCount,
            nextRetry,
            retryErrors,
            [ExecutionId]: retryId,
            [CancellationToken]: cancellationToken
        };
        const exec = fn
            ? (...args) => {
                ++context[ExecutionId].value;
                return defaultStrategy(opt, context, fn, args);
            }
            : (f) => {
                ++context[ExecutionId].value;
                return defaultStrategy(opt, context, f, undefined);
            };
        const cancel = () => {
            context.isRetrying.value = false;
            context.retryErrors.value.push(new Error("[useRetry] cancelled"));
            context.nextRetry.value = undefined;
            cancellationToken.value = true;
        };
        return {
            ...context,
            cancel,
            exec
        };
    }
    const exponentialDelay = retryNumber => {
        const delay = Math.pow(2, retryNumber) * 100;
        const randomSum = delay * 0.2 * Math.random(); // 0-20% of the delay
        return delay + randomSum;
    };
    const noDelay = () => 0;

    function useDebounce(handler, wait, options) {
        return debounce(handler, wait, options);
    }
    function debounce(func, waitMilliseconds = 50, options = {
        isImmediate: false
    }) {
        let timeoutId;
        return function (...args) {
            const context = this;
            const doLater = function () {
                timeoutId = undefined;
                if (!options.isImmediate) {
                    func.apply(context, args);
                }
            };
            const shouldCallNow = options.isImmediate && timeoutId === undefined;
            if (timeoutId !== undefined) {
                clearTimeout(timeoutId);
            }
            timeoutId = setTimeout(doLater, waitMilliseconds);
            if (shouldCallNow) {
                func.apply(context, args);
            }
        };
    }

    function useNow(options) {
        const SYNC_MS = 1000;
        const ms = options && options.refreshMs || SYNC_MS;
        const sync = options && isBoolean(options.sync) ? options.sync : true;
        const fn = options && isFunction(options.timeFn) && options.timeFn || Date.now;
        let handler = undefined;
        let timeoutHandler = undefined;
        const now = runtimeCore.ref(fn());
        const remove = () => {
            clearInterval(handler);
            clearTimeout(timeoutHandler);
        };
        /* istanbul ignore next */
        const start = isClient ? () => handler = setInterval(() => now.value = fn(), ms) : NO_OP;
        if (sync) {
            const offset = SYNC_MS - (now.value - (Math.floor(now.value / SYNC_MS) * SYNC_MS));
            timeoutHandler = setTimeout(start, offset);
        }
        else {
            start();
        }
        runtimeCore.onUnmounted(remove);
        return {
            now,
            remove
        };
    }

    function useDateNow(options) {
        const refreshMs = options && options.refreshMs || 1000;
        const sync = options && isBoolean(options.sync) ? options.sync : true;
        return useNow({
            refreshMs,
            sync,
            timeFn: Date.now
        });
    }

    function usePerformanceNow(options) {
        const refreshMs = options && options.refreshMs || 1000;
        const sync = options && isBoolean(options.sync) ? options.sync : true;
        return useNow({
            refreshMs,
            sync,
            timeFn: () => performance.now()
        });
    }

    function useEvent(el, name, listener, options) {
        const element = wrap(el);
        const remove = () => element.value.removeEventListener(name, listener);
        runtimeCore.onMounted(() => element.value.addEventListener(name, listener, options));
        runtimeCore.onUnmounted(remove);
        return remove;
    }

    function useOnMouseMove(el, options, wait) {
        const mouseX = runtimeCore.ref(0);
        const mouseY = runtimeCore.ref(0);
        let handler = (ev) => {
            mouseX.value = ev.x;
            mouseY.value = ev.y;
        };
        const [eventOptions, ms] = isNumber(options) ? [undefined, options] : [options, wait];
        if (ms) {
            handler = useDebounce(handler, wait);
        }
        const remove = useEvent(el, "mousemove", handler, eventOptions);
        return {
            mouseX,
            mouseY,
            remove
        };
    }

    function useOnResize(el, options, wait) {
        const element = wrap(el);
        const height = runtimeCore.ref(element.value && element.value.clientHeight);
        const width = runtimeCore.ref(element.value && element.value.clientWidth);
        let handler = () => {
            height.value = element.value.clientHeight;
            width.value = element.value.clientWidth;
        };
        const [eventOptions, ms] = isNumber(options) ? [undefined, options] : [options, wait];
        if (ms) {
            handler = useDebounce(handler, wait);
        }
        // resize seems only to be fired against the window
        const remove = isClient ? useEvent(window, "resize", handler, eventOptions || { passive: true }) : NO_OP;
        return {
            height,
            width,
            remove
        };
    }

    function useOnScroll(el, options, wait) {
        const element = wrap(el);
        const scrollTop = runtimeCore.ref(element.value && element.value.scrollTop);
        const scrollLeft = runtimeCore.ref(element.value && element.value.scrollLeft);
        let handler = (ev) => {
            scrollTop.value = element.value.scrollTop;
            scrollLeft.value = element.value.scrollLeft;
        };
        const [eventOptions, ms] = isNumber(options) ? [undefined, options] : [options, wait];
        if (ms) {
            handler = useDebounce(handler, wait);
        }
        const remove = useEvent(element, "scroll", handler, eventOptions);
        return {
            scrollTop,
            scrollLeft,
            remove
        };
    }

    function useFetch(options) {
        const json = runtimeCore.ref(null);
        // TODO add text = ref<string> ??
        const jsonError = runtimeCore.ref(null);
        const isJson = options ? options.isJson !== false : true;
        const parseImmediate = options ? options.parseImmediate !== false : true;
        const use = usePromise(async (request, init) => {
            const response = await fetch(request, init);
            if (isJson) {
                const pJson = response
                    .json()
                    .then(x => (json.value = x))
                    .catch(x => {
                    json.value = null;
                    jsonError.value = x;
                });
                if (parseImmediate) {
                    await pJson;
                }
            }
            return response;
        });
        const status = runtimeCore.computed(() => (use.result.value && use.result.value.status) || null);
        const statusText = runtimeCore.computed(() => (use.result.value && use.result.value.statusText) || null);
        return {
            ...use,
            json,
            jsonError,
            status,
            statusText
        };
    }

    function useWebSocket(url, protocols) {
        const supported = isClient && 'WebSocket' in window;
        let ws = null;
        const messageEvent = runtimeCore.ref();
        const errorEvent = runtimeCore.ref();
        const data = runtimeCore.ref(null);
        const isOpen = runtimeCore.ref(false);
        const isClosed = runtimeCore.ref(false);
        const errored = runtimeCore.ref(false);
        let send = NO_OP;
        let close = NO_OP;
        if (supported) {
            ws = new WebSocket(url, protocols);
            ws.addEventListener("message", x => {
                messageEvent.value = x;
                data.value = x.data;
            });
            ws.addEventListener("error", error => {
                errorEvent.value = error;
                errored.value = true;
            });
            ws.addEventListener("close", () => {
                isOpen.value = false;
                isClosed.value = true;
            });
            ws.addEventListener("open", () => {
                isOpen.value = true;
                isClosed.value = false;
            });
            send = (data) => ws.send(data);
            close = (code, reason) => {
                ws.close(code, reason);
            };
        }
        return {
            supported,
            ws,
            send,
            close,
            messageEvent,
            errorEvent,
            data,
            isOpen,
            isClosed,
            errored
        };
    }

    function useIntersectionObserver(refEl, refOptions) {
        const supported = isClient && 'IntersectionObserver' in window;
        const wrappedElement = refEl ? wrap(refEl) : undefined;
        const element = wrappedElement && (isElement(wrappedElement.value) || !wrappedElement.value)
            ? wrappedElement
            : undefined;
        const options = runtimeCore.computed(() => refOptions
            ? unwrap(refOptions)
            : !element
                ? unwrap(refEl)
                : undefined);
        const elements = runtimeCore.ref(
        // element && element.value ? [element.value] : []
        );
        const isIntersecting = runtimeCore.computed(() => elements.value.length > 0 && elements.value.every(x => x.isIntersecting));
        const handling = (entries) => {
            elements.value = entries;
        };
        let observer = runtimeCore.ref();
        if (supported) {
            runtimeCore.watch(options, options => {
                if (observer.value) {
                    observer.value.disconnect();
                }
                const opts = (options &&
                    options && {
                    root: unwrap(options.root),
                    rootMargin: unwrap(options.rootMargin),
                    threshold: unwrap(options.threshold)
                }) ||
                    undefined;
                observer.value = new IntersectionObserver(handling, opts);
                const targets = elements.value.map(x => x.target);
                targets.forEach(observer.value.observe);
            }, { deep: true });
        }
        const observe = supported ? (element) => {
            const e = unwrap(element);
            observer.value.observe(e);
        } : NO_OP;
        const unobserve = supported ? (element) => {
            const e = unwrap(element);
            observer.value.unobserve(e);
        } : NO_OP;
        const disconnect = () => observer.value.disconnect();
        // if the element is passed we should add hooks
        if (element) {
            // if value is defined it is already being observed
            if (!element.value) {
                runtimeCore.onMounted(() => {
                    if (element.value) {
                        observe(element);
                    }
                });
            }
            runtimeCore.onUnmounted(() => {
                disconnect();
            });
        }
        // // debug is still work in progress, would be nice to provide some
        // // information about the target
        // /* istanbul ignore next */
        // const debug = () => {
        //   if (elements.value.length === 0) {
        //     ('production' !== 'production') && console.warn('[IntersectionObserver] no elements provided, did you mount the component?')
        //     return;
        //   }
        //   // TODO: add border to the elements 
        // };
        return {
            supported,
            elements,
            observe,
            unobserve,
            disconnect,
            isIntersecting
        };
    }

    function useNetworkInformation() {
        const connection = isClient ?
            navigator.connection ||
                navigator.mozConnection ||
                navigator.webkitConnection : false;
        const supported = !!connection;
        const downlink = runtimeCore.ref(0);
        const downlinkMax = runtimeCore.ref(0);
        const effectiveType = runtimeCore.ref("none");
        const rtt = runtimeCore.ref(0);
        const saveData = runtimeCore.ref(false);
        const type = runtimeCore.ref("none");
        let handler = NO_OP;
        let remove = NO_OP;
        /* istanbul ignore else  */
        if (connection) {
            handler = () => {
                downlink.value = connection.downlink;
                downlinkMax.value = connection.downlinkMax;
                effectiveType.value = connection.effectiveType;
                rtt.value = connection.rtt;
                saveData.value = connection.saveData;
                type.value = connection.type;
            };
            remove = useEvent(connection, "change", handler, { passive: true });
            handler();
        }
        return {
            supported,
            downlink,
            downlinkMax,
            effectiveType,
            rtt,
            saveData,
            type,
            remove
        };
    }

    let online = undefined;
    function useOnline() {
        const supported = isClient && "onLine" in navigator;
        // not sure how to test this :/
        if (!supported) {
            online = runtimeCore.ref(false);
        }
        if (!online) {
            online = runtimeCore.ref(navigator.onLine);
            window.addEventListener("offline", () => (online.value = false), {
                passive: true
            });
            window.addEventListener("online", () => (online.value = true), {
                passive: true
            });
        }
        return {
            supported,
            online
        };
    }

    let visibility = undefined;
    let hidden = undefined;
    function usePageVisibility() {
        if (!hidden) {
            hidden = runtimeCore.ref(isClient && document.hidden);
        }
        if (!visibility) {
            if (isClient) {
                visibility = runtimeCore.ref(document.visibilityState);
                document.addEventListener("visibilitychange", () => {
                    visibility.value = document.visibilityState;
                    hidden.value = document.hidden;
                }, { passive: true }
                // true
                );
            }
            else {
                visibility = runtimeCore.ref('visible');
            }
        }
        return {
            visibility,
            hidden
        };
    }

    let language = undefined;
    let languages = undefined;
    function useLanguage() {
        if (!language) {
            language = isClient ? runtimeCore.ref(navigator.language) : runtimeCore.ref('');
        }
        if (!languages) {
            if (isClient) {
                languages = runtimeCore.ref(navigator.languages);
                const change = () => {
                    language.value = navigator.language;
                    languages.value = navigator.languages;
                };
                window.addEventListener('languagechange', change, { passive: true });
            }
            else {
                languages = runtimeCore.ref([]);
            }
        }
        return {
            language,
            languages
        };
    }

    function useBroadcastChannel(name, onBeforeClose) {
        const supported = isClient && 'BroadcastChannel' in self;
        const data = runtimeCore.ref(null);
        const messageEvent = runtimeCore.ref(null);
        const errorEvent = runtimeCore.ref(null);
        const errored = runtimeCore.ref(false);
        const isClosed = runtimeCore.ref(false);
        let send = NO_OP;
        let close = NO_OP;
        let addListener = NO_OP;
        /* istanbul ignore else  */
        if (supported) {
            const bc = new BroadcastChannel(name);
            bc.addEventListener('messageerror', (e) => {
                errorEvent.value = e;
                errored.value = true;
            }, PASSIVE_EV);
            bc.addEventListener('message', (ev) => {
                messageEvent.value = ev;
                data.value = ev.data;
            }, PASSIVE_EV);
            send = (d) => bc.postMessage(d);
            close = () => {
                bc.close();
                isClosed.value = true;
            };
            addListener = (cb, o) => {
                bc.addEventListener('message', cb, o);
                runtimeCore.onUnmounted(() => bc.removeEventListener('message', cb));
            };
            runtimeCore.onUnmounted(() => {
                onBeforeClose && onBeforeClose();
                close();
            });
        }
        return {
            supported,
            data,
            messageEvent,
            errorEvent,
            errored,
            isClosed,
            send,
            close,
            addListener,
        };
    }

    function useGeolocation(options) {
        const supported = isClient && !!navigator.geolocation;
        // used to check if the execution is lazy
        const lazy = runtimeCore.ref(options ? options.immediate === false : undefined);
        const error = runtimeCore.ref(null);
        const timestamp = runtimeCore.ref(null);
        const coords = runtimeCore.ref(null);
        const highAccuracy = runtimeCore.ref(options && options.enableHighAccuracy || null);
        // allow manual control on when the geolocation is requested
        let refresh = NO_OP;
        if (supported) {
            const setPosition = (pos) => {
                timestamp.value = pos.timestamp;
                coords.value = pos.coords;
                error.value = null;
            };
            const setError = (err) => {
                timestamp.value = Date.now();
                coords.value = null;
                error.value = err;
            };
            const clearWatch = () => lazy.value !== true && watchId && navigator.geolocation.clearWatch(watchId);
            let _currentPositionRefresh = () => navigator.geolocation.getCurrentPosition(setPosition, setError, options);
            if (lazy.value) {
                refresh = () => {
                    if (lazy.value) {
                        lazy.value = false;
                    }
                    else {
                        _currentPositionRefresh();
                    }
                };
            }
            else {
                // NOTE probably useless??
                refresh = _currentPositionRefresh;
            }
            let watchId = 0;
            runtimeCore.onMounted(() => runtimeCore.watch([highAccuracy, lazy], (a) => {
                clearWatch();
                const enableHighAccuracy = isBoolean(a[0]) ? a[0] : options ? options.enableHighAccuracy : undefined;
                watchId = navigator.geolocation.watchPosition(setPosition, setError, options ? { ...options, enableHighAccuracy } : { enableHighAccuracy });
            }, {
                lazy: lazy.value
            }));
            runtimeCore.onUnmounted(clearWatch);
        }
        return {
            supported,
            refresh,
            error,
            timestamp,
            coords,
            highAccuracy
        };
    }

    function useMatchMedia(query) {
        const supported = isClient ? 'matchMedia' in window : false;
        let mediaQueryList = undefined;
        let matches = undefined;
        let remove = NO_OP;
        if (supported) {
            mediaQueryList = runtimeCore.ref(matchMedia(query));
            matches = runtimeCore.ref(mediaQueryList.value.matches);
            const process = (e) => {
                matches.value = e.matches;
            };
            mediaQueryList.value.addEventListener("change", process, { passive: true });
            const remove = () => mediaQueryList.value.removeEventListener("change", process);
            runtimeCore.onUnmounted(remove);
        }
        else {
            mediaQueryList = runtimeCore.ref({});
            matches = runtimeCore.ref(false);
        }
        return {
            supported,
            mediaQueryList,
            matches,
            remove
        };
    }

    function useBreakpoint(breakpoints) {
        const result = {};
        const map = new Map();
        const current = runtimeCore.ref();
        let sorted = [];
        const removeMedia = [];
        for (const key in breakpoints) {
            const bp = breakpoints[key];
            if (isNumber(bp)) {
                const r = runtimeCore.ref(false);
                result[key] = r;
                map.set(bp, {
                    name: key,
                    valid: r
                });
                sorted.push(bp);
            }
            else {
                const { matches, remove } = useMatchMedia(bp);
                result[key] = matches;
                removeMedia.push(remove);
            }
        }
        sorted = sorted.sort((a, b) => b - a);
        const resize = isClient ? () => {
            const width = window.innerWidth;
            let c = undefined;
            for (let i = 0; i < sorted.length; i++) {
                const bp = sorted[i];
                const r = map.get(bp);
                r.valid.value = width >= bp;
                if (width >= bp && c === undefined) {
                    c = r.name;
                }
            }
            current.value = c;
        } : NO_OP;
        const processResize = useDebounce(resize, 10);
        const remove = isClient ? () => window.removeEventListener("resize", processResize) : NO_OP;
        runtimeCore.onMounted(() => {
            resize();
            window.addEventListener("resize", processResize, {
                passive: true
            });
        });
        runtimeCore.onUnmounted(() => {
            remove();
            removeMedia.forEach(x => x());
        });
        return {
            ...result,
            remove,
            current
        };
    }

    function useSharedRef(name, defaultValue) {
        const { addListener, send, close, supported } = useBroadcastChannel(name, () => disconnect());
        const id = Date.now();
        const master = runtimeCore.ref(false);
        const mind = runtimeCore.ref(0 /* HIVE */);
        const editable = runtimeCore.computed(() => mind.value === 1 /* MASTER */ ? master.value : true);
        // who's listening to this broadcast
        const targets = runtimeCore.ref([]);
        const data = runtimeCore.ref(defaultValue);
        // if the state was updated by an event it sets to true
        let updateState = false;
        let masterId = undefined;
        send({ type: 0 /* INIT */ });
        const ping = () => send({ type: 5 /* PING */, id });
        const disconnect = () => {
            if (targets.value.length === 0)
                return;
            if (master.value) {
                send({
                    type: 3 /* SET_MIND */,
                    mind: 1 /* MASTER */,
                    id: Math.min(...targets.value)
                });
            }
            send({
                type: 4 /* LEAVE */,
                id
            });
        };
        const setMind = (t) => {
            switch (t) {
                case 1 /* MASTER */: {
                    master.value = true;
                    break;
                }
                case 0 /* HIVE */: {
                    master.value = false;
                    break;
                }
            }
            mind.value = t;
            send({
                type: 3 /* SET_MIND */,
                id: id,
                mind: mind.value,
            });
        };
        addListener((e) => {
            switch (e.data.type) {
                case 0 /* INIT */: {
                    send({
                        type: 2 /* UPDATE */,
                        value: data.value,
                        mind: mind.value
                    });
                    break;
                }
                case 4 /* LEAVE */: {
                    const index = targets.value.indexOf(e.data.id);
                    if (index >= 0) {
                        targets.value.splice(index, 1);
                    }
                    // if master disconnects 
                    if (masterId === e.data.id && targets.value.length > 0) {
                        send({
                            type: 3 /* SET_MIND */,
                            mind: 1 /* MASTER */,
                            id: Math.min(id, ...targets.value)
                        });
                    }
                    break;
                }
                case 2 /* UPDATE */: {
                    updateState = true;
                    data.value = e.data.value;
                    mind.value = e.data.mind;
                    break;
                }
                case 3 /* SET_MIND */: {
                    mind.value = e.data.mind;
                    masterId = e.data.mind === 1 /* MASTER */ && e.data.id || undefined;
                    master.value = masterId === id;
                    if (master.value) {
                        targets.value = [];
                        ping();
                    }
                    break;
                }
                case 5 /* PING */: {
                    targets.value = [e.data.id];
                    send({
                        type: 6 /* PONG */,
                        id
                    });
                    break;
                }
                case 6 /* PONG */: {
                    targets.value.push(e.data.id);
                    break;
                }
            }
        }, PASSIVE_EV);
        ping();
        runtimeCore.watch(data, (v, o) => {
            if (updateState) {
                updateState = false;
                return;
            }
            // mind is set to MASTER and we are not master, we shouldn't update!
            if (mind.value === 1 /* MASTER */ && master.value === false) {
                updateState = true;
                data.value = o;
                return;
            }
            send({
                type: 2 /* UPDATE */,
                mind: mind.value,
                value: isObject(v) ? { ...v } : v
            });
            updateState = false;
        }, { deep: true, lazy: true });
        if (isClient) {
            window.addEventListener('unload', disconnect, PASSIVE_EV);
        }
        runtimeCore.onUnmounted(() => {
            disconnect();
            close();
        });
        return {
            supported,
            id,
            data,
            master,
            mind,
            editable,
            targets,
            ping,
            setMind,
            addListener
        };
    }
    function refShared(defaultValue, id) {
        const vm = runtimeCore.getCurrentInstance();
        const name = id ? id : vm.vnode.el;
        const { data, supported } = useSharedRef(name, defaultValue);
        return data;
    }

    const STORAGE_TEST_KEY =  ":$";
    /* istanbul ignore next */
    function isQuotaExceededError(e, storage) {
        return e instanceof DOMException && (
        // everything except Firefox
        e.code === 22 ||
            // Firefox
            e.code === 1014 ||
            // test name field too, because code might not be present
            // everything except Firefox
            e.name === 'QuotaExceededError' ||
            // Firefox
            e.name === 'NS_ERROR_DOM_QUOTA_REACHED') &&
            // acknowledge QuotaExceededError only if there's something already stored
            (storage && storage.length !== 0 || false);
    }
    // based on https://developer.mozilla.org/en-US/docs/Web/API/Web_Storage_API/Using_the_Web_Storage_API
    function storageAvailable(storage) {
        try {
            if (!storage) {
                return false;
            }
            const x = STORAGE_TEST_KEY;
            storage.setItem(x, x);
            storage.removeItem(x);
            return true;
        }
        catch (e) {
            return isQuotaExceededError(e, storage);
        }
    }
    function safeParse(serializer, value) {
        try {
            return serializer.parse(value);
        }
        catch {
            return value;
        }
    }
    let storageMap = undefined;
    function useWebStorage(type, serializer = JSON, ms = 10) {
        const storage = isClient ? window[type] : undefined;
        const supported = storageAvailable(storage);
        const remove = () => storageMap.delete(type);
        if (!storageMap) {
            storageMap = new Map();
            if (isClient) {
                window.addEventListener('storage', (e) => {
                    if (e.newValue === e.oldValue) {
                        return;
                    }
                    let webStore = storageMap.get('localStorage');
                    if (e.storageArea === window.localStorage) {
                        webStore = storageMap.get('localStorage');
                    }
                    else {
                        webStore = storageMap.get('sessionStorage');
                    }
                    if (webStore && Object.keys(webStore.$syncKeys).length > 0) {
                        if (e.key === null) {
                            webStore.clear();
                        }
                        else if (webStore.$syncKeys[e.key]) {
                            if (e.newValue === null) {
                                webStore.removeItem(e.key);
                            }
                            else {
                                webStore.updateItem(e.key, e.newValue);
                            }
                        }
                    }
                });
            }
        }
        let store = storageMap.get(type);
        let quotaError;
        if (supported && storage) {
            if (!store) {
                quotaError = runtimeCore.ref(false);
                store = {
                    $refMap: new Map(),
                    $watchHandlers: new Map(),
                    $syncKeys: {},
                    $quotaError: quotaError,
                    key: storage.key,
                    length: storage.length,
                    setSync(key, sync) {
                        if (sync) {
                            this.$syncKeys[key] = true;
                        }
                        else {
                            delete this.$syncKeys[key];
                        }
                    },
                    clear() {
                        this.$refMap.forEach((_, k) => this.removeItem(k));
                    },
                    removeItem(k) {
                        const item = this.$refMap.get(k);
                        // remove the object value if item deleted
                        if (item) {
                            item.value = undefined;
                        }
                        // clear the watch
                        const stop = this.$watchHandlers.get(k);
                        if (stop) {
                            stop();
                        }
                        delete this.$syncKeys[k];
                        this.$refMap.delete(k);
                        storage.removeItem(k);
                    },
                    getItem(k) {
                        let r = this.$refMap.get(k);
                        if (r) {
                            return r;
                        }
                        const data = storage.getItem(k);
                        if (!data) {
                            return null;
                        }
                        return this.setItem(k, safeParse(serializer, data));
                    },
                    setItem(k, v) {
                        const reference = wrap(v);
                        this.$refMap.set(k, reference);
                        const save = (key, value) => {
                            try {
                                const data = isString(value) ? value : serializer.stringify(value);
                                storage.setItem(key, data);
                            }
                            catch (e) {
                                quotaError.value = isQuotaExceededError(e, storage);
                            }
                        };
                        save(k, v);
                        const stop = runtimeCore.watch(() => reference, debounce((r) => {
                            save(k, r.value);
                        }, ms), {
                            lazy: true,
                            deep: true
                        });
                        this.$watchHandlers.set(k, stop);
                        return reference;
                    },
                    updateItem(k, data) {
                        let r = this.$refMap.get(k);
                        if (r) {
                            r.value = safeParse(serializer, data);
                        }
                    }
                };
                storageMap.set(type, store);
            }
            else {
                quotaError = store.$quotaError;
            }
        }
        else {
            quotaError = runtimeCore.ref(false);
            store = {};
        }
        return {
            supported,
            quotaError,
            store,
            remove
        };
    }

    function useLocalStorage(key, defaultValue) {
        const { supported, store } = useWebStorage('localStorage');
        let remove = NO_OP;
        let clear = NO_OP;
        let setSync = NO_OP;
        let storage = undefined;
        if (supported && store) {
            setSync = (s) => store.setSync(key, s);
            remove = () => store.removeItem(key);
            clear = () => store.clear();
            storage = store.getItem(key);
            if (!storage) {
                storage = store.setItem(key, defaultValue);
            }
        }
        else {
            storage = runtimeCore.ref(defaultValue);
        }
        return {
            supported,
            storage,
            clear,
            remove,
            setSync,
        };
    }

    function useSessionStorage(key, defaultValue) {
        const { supported, store } = useWebStorage('sessionStorage');
        let remove = NO_OP;
        let clear = NO_OP;
        let setSync = FALSE_OP;
        let storage = undefined;
        if (supported && store) {
            remove = () => store.removeItem(key);
            clear = () => store.clear();
            storage = store.getItem(key);
            if (!storage) {
                storage = store.setItem(key, defaultValue);
            }
        }
        else {
            storage = runtimeCore.ref(defaultValue);
        }
        return {
            supported,
            storage,
            clear,
            remove,
            setSync
        };
    }

    let canUseLocalStorage = undefined;
    function useStorage(key, defaultValue) {
        if (canUseLocalStorage === undefined) {
            canUseLocalStorage = useWebStorage('localStorage').supported;
        }
        return canUseLocalStorage
            ? useLocalStorage(key, defaultValue)
            : useSessionStorage(key, defaultValue);
    }

    exports.FALSE_OP = FALSE_OP;
    exports.NO_OP = NO_OP;
    exports.PASSIVE_EV = PASSIVE_EV;
    exports.debounce = debounce;
    exports.exponentialDelay = exponentialDelay;
    exports.isArray = isArray;
    exports.isBoolean = isBoolean;
    exports.isClient = isClient;
    exports.isDate = isDate;
    exports.isElement = isElement;
    exports.isFunction = isFunction;
    exports.isNumber = isNumber;
    exports.isObject = isObject;
    exports.isPromise = isPromise;
    exports.isString = isString;
    exports.isSymbol = isSymbol;
    exports.minMax = minMax;
    exports.noDelay = noDelay;
    exports.promisedTimeout = promisedTimeout;
    exports.refShared = refShared;
    exports.storageAvailable = storageAvailable;
    exports.unwrap = unwrap;
    exports.useArrayPagination = useArrayPagination;
    exports.useBreakpoint = useBreakpoint;
    exports.useBroadcastChannel = useBroadcastChannel;
    exports.useCancellablePromise = useCancellablePromise;
    exports.useDateNow = useDateNow;
    exports.useDebounce = useDebounce;
    exports.useEvent = useEvent;
    exports.useFetch = useFetch;
    exports.useGeolocation = useGeolocation;
    exports.useIntersectionObserver = useIntersectionObserver;
    exports.useLanguage = useLanguage;
    exports.useLocalStorage = useLocalStorage;
    exports.useMatchMedia = useMatchMedia;
    exports.useNetworkInformation = useNetworkInformation;
    exports.useNow = useNow;
    exports.useOnMouseMove = useOnMouseMove;
    exports.useOnResize = useOnResize;
    exports.useOnScroll = useOnScroll;
    exports.useOnline = useOnline;
    exports.usePageVisibility = usePageVisibility;
    exports.usePagination = usePagination;
    exports.usePerformanceNow = usePerformanceNow;
    exports.usePromise = usePromise;
    exports.useRetry = useRetry;
    exports.useSessionStorage = useSessionStorage;
    exports.useSharedRef = useSharedRef;
    exports.useStorage = useStorage;
    exports.useWebSocket = useWebSocket;
    exports.useWebStorage = useWebStorage;
    exports.wrap = wrap;

    return exports;

}({}, VueRuntimeCore));
